/* 
 * EDM Demo 2015
 * Educational Datamining Demonstrator for the TCS Alignment Toolbox
 * 
 * Copyright (C) 2015
 * Benjamin Paaßen
 * AG Theoretical Computer Science
 * Centre of Excellence Cognitive Interaction Technology (CITEC)
 * University of Bielefeld
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.citec.tcs.edm2015;

import de.citec.tcs.alignment.sequence.Alphabet;
import de.citec.tcs.alignment.sequence.KeywordSpecification;
import de.citec.tcs.alignment.sequence.Node;
import de.citec.tcs.alignment.sequence.NodeSpecification;
import de.citec.tcs.alignment.sequence.Sequence;
import de.citec.tcs.alignment.sequence.StringKeywordSpecification;
import de.citec.tcs.alignment.sequence.StringValue;
import de.citec.tcs.alignment.sequence.SymbolicKeywordSpecification;
import de.citec.tcs.alignment.sequence.SymbolicValue;
import de.citec.tcs.alignment.sequence.VectorialKeywordSpecification;
import de.citec.tcs.alignment.sequence.VectorialValue;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 *
 * @author Benjamin Paassen - bpaassen@techfak.uni-bielefeld.de
 */
public class FOGSequenceReader {

	public static final NodeSpecification PROGRAM_NODE_SPECIFICATION = constructProgramNodeSpecification();

	private static NodeSpecification constructProgramNodeSpecification() {
		final Alphabet typeAlphabet = new Alphabet(new String[]{
			"annotation",
			"array access",
			"array type",
			"assignment",
			"binary",
			"block",
			"break",
			"class",
			"compilation unit",
			"compound assignment",
			"do while loop",
			"empty statement",
			"enhanced for loop",
			"expression statement",
			"for loop",
			"identifier",
			"if",
			"import",
			"literal",
			"member select",
			"method",
			"method invocation",
			"modifiers",
			"new array",
			"new class",
			"parameterized type",
			"paranthesized",
			"primitive type",
			"return",
			"type cast",
			"unary",
			"variable",
			"while loop"});
		final KeywordSpecification[] keywordSpecs = {
			new SymbolicKeywordSpecification(typeAlphabet, "type"),
			new VectorialKeywordSpecification(4, "codePosition"),
			new VectorialKeywordSpecification(1, "parent"),
			new StringKeywordSpecification("scope"),
			new StringKeywordSpecification("name"),
			new StringKeywordSpecification("className"),
			new StringKeywordSpecification("returnType"),
			new StringKeywordSpecification("externalDependencies"),
			new StringKeywordSpecification("sourceCodeFile"),};
		return new NodeSpecification(keywordSpecs);
	}

	private static final Pattern NODESTART = Pattern.compile("\\s*\\d+\\s*\\{");
	private static final Pattern EDGESTART = Pattern.compile("\\s*Edges\\s*\\{");
	private static final Pattern[] KEYWORD_PATTERNS = constructKeywordPatterns();

	private static Pattern[] constructKeywordPatterns() {
		final Pattern[] keywordPatterns = new Pattern[PROGRAM_NODE_SPECIFICATION.size()];
		for (int k = 0; k < PROGRAM_NODE_SPECIFICATION.size(); k++) {
			keywordPatterns[k] = Pattern.compile("\\s*\""
					+ PROGRAM_NODE_SPECIFICATION.getKeyword(k) + "\"\\s*:\\s*"
					+ "\"([^\"]+)\"\\s*;");
		}
		return keywordPatterns;
	}

	public static Sequence readFOG(final String filePath) throws IOException {
		final File inputFile = new File(filePath);
		if (!inputFile.exists() || !inputFile.isFile()) {
			throw new FileNotFoundException("File " + inputFile.getAbsolutePath() + " could not be found!");
		}
		final Sequence seq = new Sequence(PROGRAM_NODE_SPECIFICATION);
		Node n = null;
		String line;
		final BufferedReader in = new BufferedReader(new FileReader(inputFile));
		while ((line = in.readLine()) != null) {
			if (NODESTART.matcher(line).matches()) {
				n = new Node(seq);
				seq.getNodes().add(n);
				n.setValue("sourceCodeFile", new StringValue(inputFile.getName().replaceFirst("\\.fog", ".java")));
				continue;
			}
			if (EDGESTART.matcher(line).matches()) {
				break;
			}
			for (int k = 0; k < KEYWORD_PATTERNS.length; k++) {
				final Matcher m = KEYWORD_PATTERNS[k].matcher(line);
				if (m.matches()) {
					final String content = m.group(1);
					switch (k) {
						// type
						case 0:
							final Alphabet typeAlphabet = ((SymbolicKeywordSpecification) PROGRAM_NODE_SPECIFICATION.getKeywordSpecification(0)).getAlphabet();
							try {
								n.setValue(k, new SymbolicValue(typeAlphabet, content));
							} catch (UnsupportedOperationException ex) {
								System.out.println(content);
								throw ex;
							}
							break;
						// parent
						case 2:
							n.setValue(k, new VectorialValue(Integer.parseInt(content)));
							break;
						// codePosition
						case 1:
							final String[] numberStrings = content.substring(1, content.length() - 1).split(", ");
							if(numberStrings.length != 4){
								throw new RuntimeException("unexpected code position string: " + content);
							}
							final double[] codePosition = new double[numberStrings.length];
							for(int i = 0; i < numberStrings.length; i++){
								codePosition[i] = Integer.parseInt(numberStrings[i]);
							}
							n.setValue(k, new VectorialValue(codePosition));
							break;
						default:
							n.setValue(k, new StringValue(content));
					}
					break;
				}
			}
		}
		in.close();

		return seq;
	}

}
