/* 
 * EDM Demo 2015
 * Educational Datamining Demonstrator for the TCS Alignment Toolbox
 * 
 * Copyright (C) 2015
 * Benjamin Paaßen
 * AG Theoretical Computer Science
 * Centre of Excellence Cognitive Interaction Technology (CITEC)
 * University of Bielefeld
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.citec.tcs.edm2015;

import de.citec.tcs.alignment.AlignmentPath;
import de.citec.tcs.alignment.AlignmentSpecification;
import de.citec.tcs.alignment.Operation;
import de.citec.tcs.alignment.StrictAlignmentFullAlgorithm;
import de.citec.tcs.alignment.StrictAlignmentScoreAlgorithm;
import de.citec.tcs.alignment.comparators.CharStatComparator;
import de.citec.tcs.alignment.comparators.Comparator;
import de.citec.tcs.alignment.comparators.GapComparator;
import de.citec.tcs.alignment.comparators.ReplacementComparator;
import de.citec.tcs.alignment.comparators.ReplacementCosts;
import de.citec.tcs.alignment.sequence.Alphabet;
import de.citec.tcs.alignment.sequence.Node;
import de.citec.tcs.alignment.sequence.NodeSpecification;
import de.citec.tcs.alignment.sequence.Sequence;
import de.citec.tcs.alignment.sequence.StringValue;
import de.citec.tcs.alignment.sequence.SymbolicKeywordSpecification;
import de.citec.tcs.alignment.sequence.SymbolicValue;
import de.citec.tcs.alignment.sequence.Value;
import de.citec.tcs.alignment.sequence.ValueType;
import de.citec.tcs.alignment.sequence.VectorialValue;
import de.citec.tcs.alignment.wrappers.StringEditDistance;
import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JEditorPane;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JSpinner;
import javax.swing.SpinnerNumberModel;

/**
 *
 * @author Benjamin Paassen - bpaassen(at)techfak.uni-bielefeld.de
 */
public class Step4Card extends Card implements ActionListener {

	private static final String[] usedTypes = {"variable", "literal", "assignment"};
	private static final String[] usedKeywords = {"type", "name", "className"};
	private static final Sequence[] sequences = setupSequences();
	private static final int VGAP = 20;

	private JSpinner[] weightSpinners;
	private JSpinner[][] repMatSpinners;
	private JComboBox<String> stringMetric;
	private JEditorPane alignmentPane;

	public Step4Card() {
		super("Parameters", "step4.html");
	}

	@Override
	public void fillRightPanel(JPanel superPanel) {
		superPanel.setLayout(new BorderLayout());

		final Dimension superSize = superPanel.getPreferredSize();

		final Dimension size = new Dimension(superSize.width, (int) (1.5 * superSize.height));
		// wrap everything in a scroll pane.
		final JPanel panel = new JPanel();
		panel.setLayout(new BoxLayout(panel, BoxLayout.PAGE_AXIS));

		JScrollPane scroll = new JScrollPane(panel);
		scroll.setPreferredSize(superSize);
		scroll.setMinimumSize(superSize);
		superPanel.add(scroll, BorderLayout.CENTER);

		panel.add(Box.createRigidArea(new Dimension(0, VGAP)));
		// show weights
		panel.add(new JLabel("Feature Weights"));
		{
			weightSpinners = new JSpinner[usedKeywords.length];
			for (int k = 0; k < usedKeywords.length; k++) {
				final JPanel weightPanel = new JPanel(new GridBagLayout());
				panel.add(weightPanel);
				final GridBagConstraints c = new GridBagConstraints();
				c.weightx = 0.8;
				c.gridx = 0;
				c.gridy = 0;
				c.fill = GridBagConstraints.HORIZONTAL;
				weightPanel.add(new JLabel(usedKeywords[k]), c);
				c.gridx++;
				c.weightx = 0.3;
				final SpinnerNumberModel weightModel = new SpinnerNumberModel(
						1, 0, 10, 1);
				weightSpinners[k] = new JSpinner(weightModel);
				weightPanel.add(weightSpinners[k], c);
			}
		}
		panel.add(Box.createRigidArea(new Dimension(0, VGAP)));
		// show replacement matrix.
		panel.add(new JLabel("Distances for the 'type' feature"));
		{
			JPanel repMatPanel = new JPanel(new GridBagLayout());
			panel.add(repMatPanel);
			final GridBagConstraints c2 = new GridBagConstraints();
			c2.ipadx = 10;
			c2.ipady = 10;
			c2.gridx = 0;
			c2.gridy = 0;
			c2.fill = GridBagConstraints.HORIZONTAL;
			c2.weightx = 0.2;
			c2.weighty = 0.2;

			repMatPanel.add(new JLabel("vs."), c2);
			for (int a = 0; a < usedTypes.length + 1; a++) {
				final String symbol;
				if (a < usedTypes.length) {
					symbol = usedTypes[a];
				} else {
					symbol = "-";
				}
				c2.gridx = a + 1;
				c2.gridy = 0;
				repMatPanel.add(new JLabel(symbol), c2);
				c2.gridx = 0;
				c2.gridy = a + 1;
				repMatPanel.add(new JLabel(symbol), c2);
			}

			repMatSpinners = new JSpinner[usedTypes.length + 1][usedTypes.length + 1];

			for (int i = 0; i < usedTypes.length + 1; i++) {
				for (int j = 0; j < usedTypes.length + 1; j++) {
					c2.gridx = i + 1;
					c2.gridy = j + 1;

					final double defaultValue;
					if (i == j) {
						defaultValue = 0;
					} else {
						defaultValue = 1;
					}
					final SpinnerNumberModel spinnerModel
							= new SpinnerNumberModel(defaultValue, 0, 1, 0.1);
					repMatSpinners[i][j] = new JSpinner(spinnerModel);
					repMatPanel.add(repMatSpinners[i][j], c2);
				}
			}
		}
		panel.add(Box.createRigidArea(new Dimension(0, VGAP)));
		
		// show metric select for name feature
		panel.add(new JLabel("String Metric for 'name' feature"));
		stringMetric = new JComboBox<>(new String[]{"edit distance", "character frequencies"});
		panel.add(stringMetric);
		panel.add(Box.createRigidArea(new Dimension(0, VGAP)));

		// show alignment button
		final JButton alignButton = new JButton("align");
		alignButton.addActionListener(this);
		panel.add(alignButton);
		panel.add(Box.createRigidArea(new Dimension(0, VGAP)));

		// show a JEditorPane for the result
		{
			this.alignmentPane = new JEditorPane("text", "");
			alignmentPane.setEditable(false);
			JScrollPane alignmentScroll = new JScrollPane(alignmentPane);

			final Dimension alignmentScrollSize = new Dimension((int) (0.9 * size.width), (int) (0.5 * size.height));
			alignmentScroll.setPreferredSize(alignmentScrollSize);
			alignmentScroll.setMinimumSize(alignmentScrollSize);
			panel.add(alignmentScroll);
		}
	}

	@Override
	public void actionPerformed(ActionEvent ae) {
		final NodeSpecification nodeSpec = FOGSequenceReader.PROGRAM_NODE_SPECIFICATION;
		// retrieve weighting
		final double[] w = new double[usedKeywords.length];
		double normalization = 0;
		for (int k = 0; k < usedKeywords.length; k++) {
			w[k] = (Integer) weightSpinners[k].getValue();
			normalization += w[k];
		}
		for (int k = 0; k < usedKeywords.length; k++) {
			w[k] /= normalization;
		}
		// retrieve replacement matrix.
		final Alphabet typeAlpha = ((SymbolicKeywordSpecification) nodeSpec.getKeywordSpecification("type")).getAlphabet();
		final double[][] C = new double[typeAlpha.size() + 1][typeAlpha.size() + 1];
		for (int i = 0; i < repMatSpinners.length; i++) {
			final int i2;
			if (i < usedTypes.length) {
				i2 = typeAlpha.getKeywordIndex(usedTypes[i]);
			} else {
				i2 = typeAlpha.size();
			}
			for (int j = 0; j < repMatSpinners.length; j++) {
				final int j2;
				if (j < usedTypes.length) {
					j2 = typeAlpha.getKeywordIndex(usedTypes[j]);
				} else {
					j2 = typeAlpha.size();
				}
				C[i2][j2] = (Double) repMatSpinners[i][j].getValue();
			}
		}
		// retrieve comparator for names
		final Comparator nameComparator;
		if (stringMetric.getSelectedItem().equals("edit distance")) {
			nameComparator = new EditDistanceComparator();
		} else {
			nameComparator = new CharStatComparator();
		}
		// set up comparators.
		final Comparator[] comparators = {
			new ReplacementComparator(new ReplacementCosts(typeAlpha, C)),
			nameComparator,
			new CharStatComparator()
		};
		// set up alignment specification
		final AlignmentSpecification alignSpec = new AlignmentSpecification(
				nodeSpec, usedKeywords, comparators, w);
		// set up algorithm.
		final StrictAlignmentFullAlgorithm algo = new StrictAlignmentFullAlgorithm(alignSpec);
		// calculate alignment and show it.
		final AlignmentPath path = algo.calculateAlignment(sequences[0], sequences[1]);
		alignmentPane.setText(editDistanceString(path));
	}

	private static Sequence[] setupSequences() {
		final Sequence[] seqs = new Sequence[2];
		final NodeSpecification nodeSpec = FOGSequenceReader.PROGRAM_NODE_SPECIFICATION;
		final Alphabet typeAlpha = ((SymbolicKeywordSpecification) nodeSpec.getKeywordSpecification("type")).getAlphabet();

		{
			// build first sequence. This is just one variable declaration
			// and one identifier
			seqs[0] = new Sequence(nodeSpec);
			Node n = new Node(seqs[0]);
			seqs[0].getNodes().add(n);
			n.setValue("type", new SymbolicValue(typeAlpha, "variable"));
			n.setValue("scope", new StringValue("[]"));
			n.setValue("name", new StringValue("myVar"));
			n.setValue("className", new StringValue("int"));
			n.setValue("externalDependencies", new StringValue("[int]"));

			n = new Node(seqs[0]);
			seqs[0].getNodes().add(n);
			n.setValue("type", new SymbolicValue(typeAlpha, "literal"));
			n.setValue("parent", new VectorialValue(1));
			n.setValue("scope", new StringValue("[]"));
		}

		{
			// build first sequence. This is just an additional assignment
			seqs[1] = new Sequence(nodeSpec);
			Node n = new Node(seqs[1]);
			seqs[1].getNodes().add(n);
			n.setValue("type", new SymbolicValue(typeAlpha, "variable"));
			n.setValue("scope", new StringValue("[]"));
			n.setValue("name", new StringValue("yourVar"));
			n.setValue("className", new StringValue("int"));
			n.setValue("externalDependencies", new StringValue("[int]"));

			n = new Node(seqs[1]);
			seqs[1].getNodes().add(n);
			n.setValue("type", new SymbolicValue(typeAlpha, "assignment"));
			n.setValue("scope", new StringValue("[]"));

			n = new Node(seqs[1]);
			seqs[1].getNodes().add(n);
			n.setValue("type", new SymbolicValue(typeAlpha, "literal"));
			n.setValue("parent", new VectorialValue(2));
			n.setValue("scope", new StringValue("[]"));
		}
		return seqs;

	}

	private static class EditDistanceComparator implements GapComparator<StringValue> {

		@Override
		public double delete(StringValue a) {
			return 1;
		}

		@Override
		public double insert(StringValue b) {
			return 1;
		}

		@Override
		public ValueType getType() {
			return ValueType.STRING;
		}

		@Override
		public double compare(StringValue a, StringValue b) {
			if (b == null) {
				if (a == null) {
					return 0;
				} else {
					return delete(a);
				}
			}
			if (a == null) {
				return insert(b);
			}

			final String left = a.getString();
			final String right = b.getString();
			final int M = left.length();
			final int N = right.length();
			final Sequence[] seqs = StringEditDistance.toSequences(new String[]{left, right});
			final AlignmentSpecification spec = StringEditDistance.setUpSpecification(seqs);
			final StrictAlignmentScoreAlgorithm algo = new StrictAlignmentScoreAlgorithm(spec);
			final Double distance = algo.calculateAlignment(seqs[0], seqs[1]);
			return distance * (M + N) / (double) Math.max(M, N);
		}

	}

	private static String editDistanceString(AlignmentPath path) {
		final StringBuilder buf = new StringBuilder();
		double total = 0;
		for (final Operation op : path.getOperations()) {
			String left = "-";
			switch (op.getType()) {
				case DELETION:
				case REPLACEMENT:
					left = nodeString(op.getLeft());
			}
			buf.append(left);
			if (left.length() < 16) {
				buf.append("\t\t");
			} else {
				buf.append("\t");
			}
			String right = "-";
			switch (op.getType()) {
				case INSERTION:
				case REPLACEMENT:
					right = nodeString(op.getRight());
			}
			buf.append(right);
			buf.append(System.getProperty("line.separator"));
			total += op.getWeightedLocalCost();
		}
		buf.append("overall distance = ");
		buf.append(Double.toString(total));
		return buf.toString();
	}

	private static String nodeString(final Node n) {
		final StringBuilder buf = new StringBuilder();
		buf.append('(');
		for (int k = 0; k < usedKeywords.length; k++) {
			if (k > 0) {
				buf.append(", ");
			}
			final Value val = n.getValue(usedKeywords[k]);
			if (val != null) {
				buf.append(val.toString());
			}
		}
		buf.append(')');
		return buf.toString();
	}

}
