/* 
 * EDM Demo 2015
 * Educational Datamining Demonstrator for the TCS Alignment Toolbox
 * 
 * Copyright (C) 2015
 * Benjamin Paaßen
 * AG Theoretical Computer Science
 * Centre of Excellence Cognitive Interaction Technology (CITEC)
 * University of Bielefeld
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.citec.tcs.edm2015;

import de.citec.tcs.alignment.AlignmentPath;
import de.citec.tcs.alignment.AlignmentSpecification;
import de.citec.tcs.alignment.ParallelProcessingEngine;
import de.citec.tcs.alignment.StrictAlignmentFullAlgorithm;
import de.citec.tcs.alignment.comparators.CharStatComparator;
import de.citec.tcs.alignment.comparators.Comparator;
import de.citec.tcs.alignment.comparators.L1NormComparator;
import de.citec.tcs.alignment.comparators.TrivialEditComparator;
import de.citec.tcs.alignment.parallel.ProgressReporter;
import de.citec.tcs.alignment.sequence.NodeSpecification;
import de.citec.tcs.alignment.sequence.Sequence;
import de.citec.tcs.alignment.visualization.HTMLVisualizer;
import de.citec.tcs.alignment.visualization.SourceCodeColumn;
import java.awt.Color;
import java.awt.Cursor;
import java.awt.Desktop;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.image.BufferedImage;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.File;
import java.io.IOException;
import javax.imageio.ImageIO;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JProgressBar;
import javax.swing.SwingWorker;

/**
 *
 * @author Benjamin Paassen - bpaassen(at)techfak.uni-bielefeld.de
 */
public class Step5Card extends Card implements ActionListener, PropertyChangeListener {

	private JButton visButton;
	private JProgressBar progressBar;
	private static final int VGAP = 20;

	public Step5Card() {
		super("Visualization", "step5.html");
	}

	@Override
	public void fillRightPanel(final JPanel panel) {
		panel.setLayout(new BoxLayout(panel, BoxLayout.PAGE_AXIS));

		panel.add(new JLabel("HTML Visualization"));
		panel.add(Box.createRigidArea(new Dimension(0, VGAP)));

		final JPanel visPanel = new JPanel();
		visPanel.setLayout(new BoxLayout(visPanel, BoxLayout.LINE_AXIS));

		this.visButton = new JButton("Create");
		visButton.addActionListener(this);
		visPanel.add(visButton);

		visPanel.add(Box.createRigidArea(new Dimension(0, VGAP)));

		final JButton showButton = new JButton("Show");
		showButton.addActionListener(new ActionListener() {

			@Override
			public void actionPerformed(ActionEvent ae) {
				final File indexFile = new File("visualization/index.html");
				if (!indexFile.exists()) {
					JOptionPane.showMessageDialog(panel,
							"The visualization file could not be found at " + indexFile.getAbsolutePath(),
							"Visualization Error",
							JOptionPane.ERROR_MESSAGE);
					return;
				}

				if (Desktop.isDesktopSupported()) {
					try {
						Desktop.getDesktop().browse(indexFile.toURI());
					} catch (IOException ex) {
						throw new RuntimeException(ex);
					}
				}
			}
		});
		visPanel.add(showButton);
		visPanel.setAlignmentX(0);
		panel.add(visPanel);

		this.progressBar = new JProgressBar(0, 100);
		progressBar.setValue(0);
		progressBar.setStringPainted(true);
		progressBar.setVisible(false);
		panel.add(progressBar);
		panel.add(Box.createRigidArea(new Dimension(0, VGAP)));

		panel.add(new JLabel("t-SNE Visualization"));
		panel.add(Box.createRigidArea(new Dimension(0, VGAP)));
		try {
			final BufferedImage tsneImage = ImageIO.read(new File(
					"visualization/tsne_picture.png"));
			final JLabel imageLabel = new JLabel(new ImageIcon(tsneImage));
			imageLabel.setBackground(Color.WHITE);
			imageLabel.setOpaque(true);
			panel.add(imageLabel);
		} catch (IOException ex) {
			throw new RuntimeException(ex);
		}

	}

	@Override
	public void actionPerformed(ActionEvent ae) {
		visButton.setEnabled(false);
		progressBar.setVisible(true);
		setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
		final VisualizationTask task = new VisualizationTask();
		task.addPropertyChangeListener(this);
		task.execute();
	}

	public static final AlignmentSpecification alignSpec = constructAlignmentSpecification();

	private static AlignmentSpecification constructAlignmentSpecification() {
		// set up comparators
		final NodeSpecification nodeSpec = FOGSequenceReader.PROGRAM_NODE_SPECIFICATION;
		final String[] keywords = {"type",
			"codePosition", "parent", "scope",
			"name", "className", "returnType", "externalDependencies"};
		final Comparator[] comparators = {
			new TrivialEditComparator(),
			new L1NormComparator(4),
			new L1NormComparator(1),
			new ScopeComparator(),
			new CharStatComparator(),
			new CharStatComparator(),
			new CharStatComparator(),
			new CharStatComparator(),};
		// set up spec.
		final AlignmentSpecification alignSpec = new AlignmentSpecification(
				nodeSpec, keywords, comparators);
		return alignSpec;
	}

	private class VisualizationTask extends SwingWorker<Void, Void>
			implements ProgressReporter {

		private static final int totalEpochs = 2;
		private static final int epochLength = 100 / totalEpochs;
		private int currentEpoch = 0;
		private int totalSteps = 0;
		private int step = 0;

		@Override
		protected Void doInBackground() throws Exception {
			// create all pairwise alignments for the dataset.
			final Sequence[] data = Step2Card.programSequences.values().
					toArray(new Sequence[0]);
			// set up algorithm.
			final StrictAlignmentFullAlgorithm algo
					= new StrictAlignmentFullAlgorithm(alignSpec);
			// do parallel processing.
			final ParallelProcessingEngine<AlignmentPath> engine
					= new ParallelProcessingEngine<>(algo, data);
			engine.setReporter(this);
			engine.setFull();
			engine.calculate();
			// retrieve the result.
			final AlignmentPath[][] D = engine.getResultMatrix();
			// visualize it.
			final HTMLVisualizer vis = new HTMLVisualizer();
			vis.setReporter(this);
			final SourceCodeColumn srcCol = new SourceCodeColumn(
					"sourceCode", "codePosition", "sourceCodeFile", "programs");
			srcCol.setTabSize(4);
			vis.getAdditionalColumns().add(srcCol);
			vis.visualizeMatrix(D,
					Step2Card.programSequences.keySet().toArray(new String[0]),
					new File("visualization"));
			setProgress(100);
			return null;
		}

		@Override
		public void setTotalSteps(int totalSteps) {
			this.totalSteps = totalSteps;
		}

		@Override
		public int getTotalSteps() {
			return totalSteps;
		}

		@Override
		public void resetSteps() {
			this.step = 0;
			currentEpoch++;
		}

		@Override
		public void incrementSteps() {
			this.step++;
			final double progress = epochLength * (currentEpoch - 1) + (double) step / (double) totalSteps * epochLength;
			setProgress((int) Math.round(progress));
		}

		@Override
		public int getSteps() {
			return step;
		}
	}

	@Override
	public void propertyChange(PropertyChangeEvent evt) {
		if ("progress" == evt.getPropertyName()) {
			int progress = (Integer) evt.getNewValue();
			progressBar.setValue(progress);
			if (progress == 100) {
				visButton.setEnabled(true);
				setCursor(null); //turn off the wait cursor
				progressBar.setVisible(false);
			}
		}
	}

}
