/* 
 * EDM Demo 2015
 * Educational Datamining Demonstrator for the TCS Alignment Toolbox
 * 
 * Copyright (C) 2015
 * Benjamin Paaßen
 * AG Theoretical Computer Science
 * Centre of Excellence Cognitive Interaction Technology (CITEC)
 * University of Bielefeld
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.citec.tcs.edm2015;

import com.xeiam.xchart.Chart;
import com.xeiam.xchart.ChartBuilder;
import com.xeiam.xchart.StyleManager.ChartTheme;
import com.xeiam.xchart.StyleManager.ChartType;
import com.xeiam.xchart.XChartPanel;
import de.citec.tcs.alignment.AlignmentSpecification;
import de.citec.tcs.alignment.ParallelProcessingEngine;
import de.citec.tcs.alignment.StrictAlignmentFullAlgorithm;
import de.citec.tcs.alignment.StrictAlignmentScoreAlgorithm;
import de.citec.tcs.alignment.comparators.TrivialEditComparator;
import de.citec.tcs.alignment.learning.LMNNClassifier;
import de.citec.tcs.alignment.learning.LMNNGradientCalculator;
import de.citec.tcs.alignment.parallel.MatrixEngine;
import de.citec.tcs.alignment.sequence.Sequence;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;
import java.util.Random;
import java.util.TreeSet;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.SwingWorker;

/**
 *
 * @author Benjamin Paassen - bpaassen(at)techfak.uni-bielefeld.de
 */
public class Step6Card extends Card implements ActionListener {

	private JButton learnButton;
	private MetricLearningWorker learnTask;
	private JTextArea logPane;
	private XChartPanel chartPanel;
	private static final int VGAP = 20;
	private static final double MISMATCH_PANELTY = 0.2;

	public Step6Card() {
		super("Metric Learning", "step6.html");
	}

	@Override
	public void fillRightPanel(JPanel superPanel) {

		final Dimension superSize = superPanel.getPreferredSize();

		final Dimension size = new Dimension(superSize.width, (int) (1.5 * superSize.height));
		// wrap everything in a scroll pane.
		final JPanel panel = new JPanel();
		panel.setLayout(new BoxLayout(panel, BoxLayout.PAGE_AXIS));

		final JScrollPane scroll = new JScrollPane(panel);
		scroll.setPreferredSize(superSize);
		scroll.setMinimumSize(superSize);
		superPanel.add(scroll);

		// add the button for metric learning.
		panel.add(Box.createRigidArea(new Dimension(0, VGAP)));
		learnButton = new JButton("Start Metric Learning");
		learnButton.addActionListener(this);
		panel.add(learnButton);
		panel.add(Box.createRigidArea(new Dimension(0, VGAP)));

		// add the weight chart.
		{
			final Chart weightChart = new ChartBuilder().chartType(ChartType.Bar).
				theme(ChartTheme.Matlab).
				width((int) (0.9 * size.width)).height((int) (0.45 * size.height)).
				title("Weighting").build();

			final List<String> labels = Arrays.asList(Step5Card.alignSpec.getKeywords());
			final ArrayList<Number> startWeights = new ArrayList<>();
			for (final double w : Step5Card.alignSpec.getWeighting()) {
				startWeights.add(w);
			}

			weightChart.addSeries("Weighting", labels, startWeights);
			weightChart.getStyleManager().setLegendVisible(false);
			weightChart.getStyleManager().setXAxisLabelRotation(90);
			chartPanel = new XChartPanel(weightChart);
			panel.add(chartPanel);
		}
		// add the log
		panel.add(Box.createRigidArea(new Dimension(0, VGAP)));
		panel.add(new JLabel("Progress"));
		panel.add(Box.createRigidArea(new Dimension(0, VGAP)));

		this.logPane = new JTextArea(20, 30);
		logPane.setMargin(new Insets(5, 5, 5, 5));
		logPane.setEditable(false);
		JScrollPane logScroll = new JScrollPane(logPane);

//		final Dimension logScrollSize = new Dimension((int) (0.9 * size.width), (int) (0.5 * size.height));
//		logScroll.setPreferredSize(logScrollSize);
//		logScroll.setMinimumSize(logScrollSize);
		panel.add(logScroll);
	}

	@Override
	public void actionPerformed(ActionEvent e) {

		if (learnTask == null) {
			logPane.setText("");

			this.learnTask = new MetricLearningWorker();
			learnTask.execute();

			learnButton.setText("Stop Metric Learning");
			setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
		} else {
			this.learnTask.cancel(true);

			learnButton.setText("Start Metric Learning");
			setCursor(null);
			learnTask = null;
		}
	}

	private static final int epochs = 5;
	private static final int K = 3;
	private static final double margin = 0.01;
	private static final int P = 3;
	private static final double eta = (double) P / (double) (Step2Card.programSequences.size() * (P - 1) * K * 150);

	private static Pattern labelPattern = Pattern.compile("[AB]{3}");

	private class MetricLearningWorker extends SwingWorker<Void, MetricLearningMessage> {

		@Override
		protected Void doInBackground() throws Exception {
			// create labels.
			final int[] labels = new int[Step2Card.programSequences.size()];
			final int N = labels.length;
			final int L;
			{
				final TreeSet<String> labelMapping = new TreeSet<>();
				final ArrayList<String> labelStrings = new ArrayList<>();

				for (final String fileName : Step2Card.programSequences.keySet()) {
					final Matcher m = labelPattern.matcher(fileName);
					m.find();
					labelStrings.add(m.group());
					labelMapping.add(m.group());
				}
				L = labelMapping.size();

				for (int i = 0; i < N; i++) {
					labels[i] = labelMapping.headSet(labelStrings.get(i)).size();
				}
			}

			// distribute in training and test set.
			final Sequence[] sequences = Step2Card.programSequences.values().toArray(new Sequence[0]);
			final ArrayList<Integer> trainingIndices = new ArrayList<>();
			final ArrayList<Integer> testIndices = new ArrayList<>();
			{
				final Random rand = new Random();
				// for each label, we want to select (N/L)/P sequences at random.
				for (int l = 0; l < L; l++) {
					final ArrayList<Integer> labelSeqs = new ArrayList<>();
					for (int i = 0; i < N; i++) {
						if (labels[i] == l) {
							labelSeqs.add(i);
						}
					}
					for (int i = 0; i < N / (L * P); i++) {
						final int test_i = labelSeqs.remove(rand.nextInt(labelSeqs.size()));
						testIndices.add(test_i);
					}
					trainingIndices.addAll(labelSeqs);
				}
			}

			// clone an AlignmentSpecification for training
			final AlignmentSpecification alignSpec = new AlignmentSpecification(Step5Card.alignSpec);
			// lower the cost for the 'type' keyword to reduce confusion due to
			// scaling effects
			final TrivialEditComparator comp = (TrivialEditComparator) alignSpec.getComparator(0);
			comp.setMismatchCost(MISMATCH_PANELTY);
			comp.setDeletionCost(MISMATCH_PANELTY);
			comp.setInsertionCost(MISMATCH_PANELTY);

			// set up algorithms
			final StrictAlignmentScoreAlgorithm scoreAlgo
				= new StrictAlignmentScoreAlgorithm(alignSpec);
			final StrictAlignmentFullAlgorithm fullAlgo
				= new StrictAlignmentFullAlgorithm(alignSpec);

			final Sequence[] trainingSequences = select(new Sequence[trainingIndices.size()], sequences, trainingIndices);
			final int[] trainingLabels = select(labels, trainingIndices);
			final int[] testLabels = select(labels, testIndices);

			// create parallel processing engines for intra-training distances
			// and for training to test distances.
			final ParallelProcessingEngine<Double> trainEngine = new ParallelProcessingEngine<>(scoreAlgo, trainingSequences);
			trainEngine.setReporter(null);
			trainEngine.setFull();
			final ParallelProcessingEngine<Double> testEngine = new ParallelProcessingEngine<>(scoreAlgo, sequences);
			testEngine.setReporter(null);
			for (final int i : testIndices) {
				for (final int j : trainingIndices) {
					testEngine.addTask(new MatrixEngine.MatrixCoordinate(i, j));
				}
			}

			// set up classifier
			final LMNNClassifier classifier = new LMNNClassifier(select(labels, trainingIndices));
			classifier.setK(K);
			classifier.setMargin(margin);

			// set up gradient calculator
			final LMNNGradientCalculator gradCalc = new LMNNGradientCalculator(trainingSequences, trainingLabels, fullAlgo);
			gradCalc.setK(K);
			gradCalc.setMargin(margin);
			gradCalc.setReporter(null);

			// check initial test accuracy.
			testEngine.calculate();
			double[][] D_test = getTrainToTestDistances(testEngine.getDoubleResultMatrix(), trainingIndices, testIndices);

			double test_acc = classifier.calculateTestAccuracy(testLabels, D_test);
			publish(new AccuracyMessage(true, test_acc, 0));

			for (int e = 1; e <= epochs; e++) {
				// calculate training distances.
				trainEngine.calculate();
				final double[][] D_train = trainEngine.getDoubleResultMatrix();
				// report training accuracy.
				final double train_acc = classifier.calculateTrainingAccuracy(D_train);
				if (isCancelled()) {
					return null;
				}
				publish(new AccuracyMessage(false, train_acc, e - 1));
				// calculate gradient.
				final double[] grad = gradCalc.calculateWeightGradient(D_train);
				double[] w = alignSpec.getWeighting();
				// do the gradient step.
				double Z = 0;
				for (int k = 0; k < w.length; k++) {
					w[k] -= eta * grad[k];
					if (w[k] < 0) {
						w[k] = 0;
					}
					Z += w[k];
				}
				// renormalize
				for (int k = 0; k < w.length; k++) {
					w[k] /= Z;
				}
				if (isCancelled()) {
					return null;
				}
				// report new weights.
				publish(new WeightMessage(w, e));
				// set new weighting
				alignSpec.setWeighting(w);
				// report new test accuracy.
				testEngine.calculate();
				D_test = getTrainToTestDistances(testEngine.getDoubleResultMatrix(), trainingIndices, testIndices);
				test_acc = classifier.calculateTestAccuracy(testLabels, D_test);
				if (isCancelled()) {
					return null;
				}
				publish(new AccuracyMessage(true, test_acc, e));
			}

			// report last training accuracy	
			final double[][] D_train = trainEngine.getDoubleResultMatrix();
			// report training accuracy.
			final double train_acc = classifier.calculateTrainingAccuracy(D_train);
			publish(new AccuracyMessage(false, train_acc, epochs));

			// report end message.
			publish(new EndMessage(epochs));
			return null;
		}

		@Override
		protected void process(List<MetricLearningMessage> msgs) {
			// publish all results in text form:
			WeightMessage lastWeight = null;
			for (final MetricLearningMessage msg : msgs) {
				if (msg instanceof AccuracyMessage) {
					final AccuracyMessage accMsg = (AccuracyMessage) msg;
					final StringBuilder text = new StringBuilder();
					if (accMsg.test) {
						text.append("Test");
					} else {
						text.append("Training");
					}
					text.append(" accuracy for epoch ");
					text.append(Integer.toString(accMsg.epoch));
					text.append(" : ");
					text.append(Double.toString(accMsg.accuracy * 100));
					text.append("%.\n");
					logPane.append(text.toString());
				} else if (msg instanceof WeightMessage) {
					final WeightMessage weightMsg = (WeightMessage) msg;
					final StringBuilder text = new StringBuilder();
					text.append("Weights for epoch ");
					text.append(Integer.toString(weightMsg.epoch));
					text.append(" : ");
					text.append(Arrays.toString(weightMsg.weights));
					text.append(".\n");
					logPane.append(text.toString());
					lastWeight = weightMsg;
				} else {
					setCursor(null);
					learnButton.setText("Start Metric Learning");
					learnTask = null;
				}
			}

			if (lastWeight != null) {
				final ArrayList<Number> yData = new ArrayList<>();
				for (final double w : lastWeight.weights) {
					yData.add(w);
				}
				chartPanel.updateSeries("Weighting", Arrays.asList(Step5Card.alignSpec.getKeywords()), yData);
			}
		}

	}

	private static class MetricLearningMessage {

		public final int epoch;

		public MetricLearningMessage(int epoch) {
			this.epoch = epoch;
		}

	}

	private static class WeightMessage extends MetricLearningMessage {

		public final double[] weights;

		public WeightMessage(double[] weights, int epoch) {
			super(epoch);
			this.weights = weights;
		}

	}

	private static class AccuracyMessage extends MetricLearningMessage {

		public final boolean test;
		public final double accuracy;

		public AccuracyMessage(boolean test, double accuracy, int epoch) {
			super(epoch);
			this.test = test;
			this.accuracy = accuracy;
		}

	}

	private static class EndMessage extends MetricLearningMessage {

		public EndMessage(int epoch) {
			super(epoch);
		}

	}

	private static <X> X[] select(X[] out, X[] actual, Collection<Integer> indices) {
		int k = 0;
		for (final int idx : indices) {
			out[k] = actual[idx];
			k++;
		}
		return out;
	}

	private static int[] select(int[] actual, Collection<Integer> indices) {
		final int[] out = new int[indices.size()];
		int k = 0;
		for (final int idx : indices) {
			out[k] = actual[idx];
		}
		return out;
	}

	private static double[][] getTrainToTestDistances(double[][] D, Collection<Integer> trainingIndices, Collection<Integer> testIndices) {
		final double[][] D_train_to_test = new double[testIndices.size()][trainingIndices.size()];
		int k = 0;
		for (final int i : testIndices) {
			int l = 0;
			for (final int j : trainingIndices) {
				D_train_to_test[k][l] = D[i][j];
				l++;
			}
			k++;
		}
		return D_train_to_test;
	}

}
